@@ -1,11 +1,29 @@
-from typing import Optional, Type
+# MIT License
+
+# Copyright (c) 2023 Replicable-MARL
+
+# Permission is hereby granted, free of charge, to any person obtaining a copy
+# of this software and associated documentation files (the "Software"), to deal
+# in the Software without restriction, including without limitation the rights
+# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+# copies of the Software, and to permit persons to whom the Software is
+# furnished to do so, subject to the following conditions:
+#
+# The above copyright notice and this permission notice shall be included in all
+# copies or substantial portions of the Software.
+#
+# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+# SOFTWARE.
+
+from typing import Type
 from ray.rllib.agents.ddpg.ddpg import DDPGTrainer, DEFAULT_CONFIG as DDPG_DEFAULT_CONFIG
 from ray.rllib.agents.ddpg.ddpg_torch_policy import DDPGTorchPolicy
-import numpy as np
 import copy
-from typing import List, Dict, Optional
-from gym.spaces import Box
-import logging
 import gym
 from typing import Tuple
 from ray.rllib.agents.dqn.dqn_tf_policy import PRIO_WEIGHTS
@@ -13,14 +31,12 @@ from ray.rllib.models.action_dist import ActionDistribution
 from ray.rllib.models.torch.torch_action_dist import TorchDeterministic, TorchDirichlet
 from ray.rllib.models.tf.tf_action_dist import Deterministic
 from ray.rllib.policy.policy import Policy
-from ray.rllib.models import ModelCatalog, MODEL_DEFAULTS
+from ray.rllib.models import MODEL_DEFAULTS
 from ray.rllib.utils.spaces.simplex import Simplex
 from ray.rllib.utils.torch_ops import huber_loss, l2_loss, sequence_mask
-from ray.rllib.utils.typing import TrainerConfigDict, TensorType, ModelInputDict
+from ray.rllib.utils.typing import TrainerConfigDict, ModelInputDict
 from ray.rllib.models.modelv2 import ModelV2
 from ray.rllib.agents.ddpg.ddpg_torch_model import DDPGTorchModel
-from ray.rllib.utils import override, force_list
-from ray.rllib.policy.sample_batch import SampleBatch
 from ray.rllib.policy.view_requirement import ViewRequirement
 from ray.rllib.utils.framework import try_import_torch
 from ray.rllib.utils.typing import ModelConfigDict, TensorType
@@ -29,101 +45,22 @@ from ray.rllib.execution.replay_buffer import *
 
 from marllib.marl.algos.utils.episode_execution_plan import episode_execution_plan
 
-
 torch, nn = try_import_torch()
 
 
-def build_rnnddpg_models(policy, observation_space, action_space, config):
-    num_outputs = int(np.product(observation_space.shape))
-
-    policy_model_config = MODEL_DEFAULTS.copy()
-    policy_model_config.update(config["policy_model"])
-    q_model_config = MODEL_DEFAULTS.copy()
-    q_model_config.update(config["Q_model"])
-
-    policy.model = ModelCatalog.get_model_v2(
-        obs_space=observation_space,
-        action_space=action_space,
-        num_outputs=num_outputs,
-        model_config=config["model"],
-        framework=config["framework"],
-        default_model=DDPG_RNN_TorchModel,
-        name="rnnddpg_model",
-        policy_model_config=policy_model_config,
-        q_model_config=q_model_config,
-        twin_q=config["twin_q"],
-        add_layer_norm=(policy.config["exploration_config"].get("type") ==
-                        "ParameterNoise"),
-    )
-
-    policy.target_model = ModelCatalog.get_model_v2(
-        obs_space=observation_space,
-        action_space=action_space,
-        num_outputs=num_outputs,
-        model_config=config["model"],
-        framework=config["framework"],
-        default_model=DDPG_RNN_TorchModel,
-        name="rnnddpg_model",
-        policy_model_config=policy_model_config,
-        q_model_config=q_model_config,
-        twin_q=config["twin_q"],
-        add_layer_norm=(policy.config["exploration_config"].get("type") ==
-                        "ParameterNoise"),
-    )
-
-    return policy.model
-
-
-def build_rnnddpg_models_and_action_dist(
-        policy: Policy, obs_space: gym.spaces.Space,
-        action_space: gym.spaces.Space,
-        config: TrainerConfigDict) -> Tuple[ModelV2, ActionDistribution]:
-    model = build_rnnddpg_models(policy, obs_space, action_space, config)
-
-    assert model.get_initial_state() != [], \
-        "RNNDDPG requires its model to be a recurrent one!"
-
-    if isinstance(action_space, Simplex):
-        return model, TorchDirichlet
-    else:
-        return model, TorchDeterministic
-
-
-def action_distribution_fn(policy: Policy,
-                           model: ModelV2,
-                           input_dict: ModelInputDict,
-                           *,
-                           explore=True,
-                           is_training=False,
-                           state_batches=None,
-                           seq_lens=None,
-                           prev_action_batch=None,
-                           prev_reward_batch=None,
-                           timestep=None,
-                           **kwargs):
-    # modify input output change
-    model_out, state_in = model(input_dict, state_batches, seq_lens)
-
-    states_in = model.select_state(state_in, ["policy", "q", "twin_q"])
-
-    distribution_inputs, policy_state_out = \
-        model.get_policy_output(model_out, states_in["policy"], seq_lens)
-    _, q_state_out = model.get_q_values(model_out, states_in["q"], seq_lens)
-    if model.twin_q_model:
-        _, twin_q_state_out = \
-            model.get_twin_q_values(model_out, states_in["twin_q"], seq_lens)
-    else:
-        twin_q_state_out = []
-
-    states_out = policy_state_out + q_state_out + twin_q_state_out
-
-    return distribution_inputs, (TorchDeterministic
-                                 if policy.config["framework"] == "torch" else
-                                 Deterministic), states_out
-
-
 def ddpg_actor_critic_loss(policy: Policy, model: ModelV2, _,
                            train_batch: SampleBatch) -> TensorType:
+    """Constructs the loss for DDPG Objective.
+    Args:
+        policy (Policy): The Policy to calculate the loss for.
+        model (ModelV2): The Model to calculate the loss for.
+        dist_class (Type[ActionDistribution]: The action distr. class.
+        train_batch (SampleBatch): The training data.
+
+    Returns:
+        Union[TensorType, List[TensorType]]: A single loss tensor or a list
+            of loss tensors.
+    """
     target_model = policy.target_models[model]
 
     i = 0
@@ -157,21 +94,14 @@ def ddpg_actor_critic_loss(policy: Policy, model: ModelV2, _,
         "prev_rewards": train_batch[SampleBatch.REWARDS],
     }
 
-    # model_out_tp1, state_in_tp1 = model(
-    #     input_dict_next, state_batches, seq_lens)
-    # states_in_tp1 = model.select_state(state_in_tp1, ["policy", "q", "twin_q"])
-
     target_model_out_tp1, target_state_in_tp1 = target_model(
         input_dict_next, state_batches, seq_lens)
     target_states_in_tp1 = target_model.select_state(target_state_in_tp1,
                                                      ["policy", "q", "twin_q"])
 
     # Policy network evaluation.
-    # prev_update_ops = set(tf1.get_collection(tf.GraphKeys.UPDATE_OPS))
     policy_t = model.get_policy_output(
         model_out_t, states_in_t["policy"], seq_lens)[0]
-    # policy_batchnorm_update_ops = list(
-    #    set(tf1.get_collection(tf.GraphKeys.UPDATE_OPS)) - prev_update_ops)
 
     policy_tp1 = target_model.get_policy_output(
         target_model_out_tp1, target_states_in_tp1["policy"], seq_lens)[0]
@@ -201,7 +131,6 @@ def ddpg_actor_critic_loss(policy: Policy, model: ModelV2, _,
         policy_tp1_smoothed = policy_tp1
 
     # Q-net(s) evaluation.
-    # prev_update_ops = set(tf1.get_collection(tf.GraphKeys.UPDATE_OPS))
     # Q-values for given actions & observations in given current
     q_t = model.get_q_values(
         model_out_t, states_in_t["q"], seq_lens, train_batch[SampleBatch.ACTIONS])[0]
@@ -214,8 +143,6 @@ def ddpg_actor_critic_loss(policy: Policy, model: ModelV2, _,
     if twin_q:
         twin_q_t = model.get_twin_q_values(model_out_t, states_in_t["twin_q"], seq_lens,
                                            train_batch[SampleBatch.ACTIONS])[0]
-    # q_batchnorm_update_ops = list(
-    #     set(tf1.get_collection(tf.GraphKeys.UPDATE_OPS)) - prev_update_ops)
 
     # Target q-net(s) evaluation.
     q_tp1 = target_model.get_q_values(
@@ -308,8 +235,97 @@ def ddpg_actor_critic_loss(policy: Policy, model: ModelV2, _,
     return actor_loss, critic_loss
 
 
-class DDPG_RNN_TorchModel(DDPGTorchModel):
-    """Extension of standard DDPGTorchModel for RNNDDPG.
+def build_iddpg_models(policy, observation_space, action_space, config):
+    num_outputs = int(np.product(observation_space.shape))
+
+    policy_model_config = MODEL_DEFAULTS.copy()
+    policy_model_config.update(config["policy_model"])
+    q_model_config = MODEL_DEFAULTS.copy()
+    q_model_config.update(config["Q_model"])
+
+    policy.model = ModelCatalog.get_model_v2(
+        obs_space=observation_space,
+        action_space=action_space,
+        num_outputs=num_outputs,
+        model_config=config["model"],
+        framework=config["framework"],
+        default_model=IDDPGTorchModel,
+        name="iddpg_model",
+        policy_model_config=policy_model_config,
+        q_model_config=q_model_config,
+        twin_q=config["twin_q"],
+        add_layer_norm=(policy.config["exploration_config"].get("type") ==
+                        "ParameterNoise"),
+    )
+
+    policy.target_model = ModelCatalog.get_model_v2(
+        obs_space=observation_space,
+        action_space=action_space,
+        num_outputs=num_outputs,
+        model_config=config["model"],
+        framework=config["framework"],
+        default_model=IDDPGTorchModel,
+        name="iddpg_model",
+        policy_model_config=policy_model_config,
+        q_model_config=q_model_config,
+        twin_q=config["twin_q"],
+        add_layer_norm=(policy.config["exploration_config"].get("type") ==
+                        "ParameterNoise"),
+    )
+
+    return policy.model
+
+
+def build_iddpg_models_and_action_dist(
+        policy: Policy, obs_space: gym.spaces.Space,
+        action_space: gym.spaces.Space,
+        config: TrainerConfigDict) -> Tuple[ModelV2, ActionDistribution]:
+    model = build_iddpg_models(policy, obs_space, action_space, config)
+
+    assert model.get_initial_state() != [], \
+        "IDDPG requires its model to be a recurrent one!"
+
+    if isinstance(action_space, Simplex):
+        return model, TorchDirichlet
+    else:
+        return model, TorchDeterministic
+
+
+def action_distribution_fn(policy: Policy,
+                           model: ModelV2,
+                           input_dict: ModelInputDict,
+                           *,
+                           explore=True,
+                           is_training=False,
+                           state_batches=None,
+                           seq_lens=None,
+                           prev_action_batch=None,
+                           prev_reward_batch=None,
+                           timestep=None,
+                           **kwargs):
+    # modify input output change
+    model_out, state_in = model(input_dict, state_batches, seq_lens)
+
+    states_in = model.select_state(state_in, ["policy", "q", "twin_q"])
+
+    distribution_inputs, policy_state_out = \
+        model.get_policy_output(model_out, states_in["policy"], seq_lens)
+    _, q_state_out = model.get_q_values(model_out, states_in["q"], seq_lens)
+    if model.twin_q_model:
+        _, twin_q_state_out = \
+            model.get_twin_q_values(model_out, states_in["twin_q"], seq_lens)
+    else:
+        twin_q_state_out = []
+
+    states_out = policy_state_out + q_state_out + twin_q_state_out
+
+    return distribution_inputs, (TorchDeterministic
+                                 if policy.config["framework"] == "torch" else
+                                 Deterministic), states_out
+
+
+class IDDPGTorchModel(DDPGTorchModel):
+    """Extension of standard DDPGTorchModel for IDDPG.
 
     Data flow:
         obs -> forward() -> model_out
@@ -495,7 +511,6 @@ class DDPG_RNN_TorchModel(DDPGTorchModel):
         out, state_out = net(model_out, state_in, seq_lens)
         return out, state_out
 
-
     @override(DDPGTorchModel)
     def get_q_values(self,
                      model_out: TensorType,
@@ -555,7 +570,7 @@ class DDPG_RNN_TorchModel(DDPGTorchModel):
 
 logger = logging.getLogger(__name__)
 
-DDPG_RNN_DEFAULT_CONFIG = DDPGTrainer.merge_trainer_configs(
+IDDPG_DEFAULT_CONFIG = DDPGTrainer.merge_trainer_configs(
     DDPG_DEFAULT_CONFIG,
     {
         "batch_mode": "complete_episodes",
@@ -578,23 +593,24 @@ def validate_config(config: TrainerConfigDict) -> None:
         config["burn_in"] + config["model"]["max_seq_len"]
 
 
-DDPGRNNTorchPolicy = DDPGTorchPolicy.with_updates(
-    name="RNNDDPGTorchPolicy",
-    get_default_config=lambda: DDPG_RNN_DEFAULT_CONFIG,
+IDDPGTorchPolicy = DDPGTorchPolicy.with_updates(
+    name="IDDPGTorchPolicy",
+    get_default_config=lambda: IDDPG_DEFAULT_CONFIG,
     action_distribution_fn=action_distribution_fn,
-    make_model_and_action_dist=build_rnnddpg_models_and_action_dist,
+    make_model_and_action_dist=build_iddpg_models_and_action_dist,
     loss_fn=ddpg_actor_critic_loss,
 )
 
 
 def get_policy_class(config: TrainerConfigDict) -> Optional[Type[Policy]]:
     if config["framework"] == "torch":
-        return DDPGRNNTorchPolicy
+        return IDDPGTorchPolicy
 
-DDPGRNNTrainer = DDPGTrainer.with_updates(
+
+IDDPGTrainer = DDPGTrainer.with_updates(
     name="IDDPGTrainer",
-    default_config=DDPG_RNN_DEFAULT_CONFIG,
-    default_policy=DDPGRNNTorchPolicy,
+    default_config=IDDPG_DEFAULT_CONFIG,
+    default_policy=None,
     get_policy_class=get_policy_class,
     validate_config=validate_config,
     allow_unknown_subkeys=["Q_model", "policy_model"],
