@@ -1,19 +1,36 @@
+# MIT License
+
+# Copyright (c) 2023 Replicable-MARL
+
+# Permission is hereby granted, free of charge, to any person obtaining a copy
+# of this software and associated documentation files (the "Software"), to deal
+# in the Software without restriction, including without limitation the rights
+# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+# copies of the Software, and to permit persons to whom the Software is
+# furnished to do so, subject to the following conditions:
+#
+# The above copyright notice and this permission notice shall be included in all
+# copies or substantial portions of the Software.
+#
+# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+# SOFTWARE.
+
 import tree  # pip install dm_tree
 import torch
 import torch.nn as nn
-from gym.spaces import Tuple as Gym_Tuple, Discrete, Dict as Gym_Dict
-
+from gym.spaces import Dict as Gym_Dict
 from ray.rllib.policy.torch_policy import TorchPolicy
 from ray.rllib.policy.policy import Policy
 from ray.rllib.models.modelv2 import _unpack_obs
 from ray.rllib.agents.qmix.model import RNNModel, _get_size
 from ray.rllib.execution.replay_buffer import *
-from ray.rllib.agents.trainer import Trainer
-from ray.rllib.evaluation.worker_set import WorkerSet
 from ray.rllib.models.torch.torch_action_dist import TorchCategorical
 from ray.rllib.utils.metrics.learner_info import LEARNER_STATS_KEY
-from ray.rllib.utils.typing import TrainerConfigDict
-from ray.util.iter import LocalIterator
 from ray.rllib.models.catalog import ModelCatalog
 from ray.rllib.policy.sample_batch import SampleBatch
 from ray.rllib.agents.qmix.qmix_policy import _mac, _validate, _unroll_mac
@@ -21,10 +38,11 @@ from ray.rllib.agents.dqn.dqn import GenericOffPolicyTrainer
 from ray.rllib.agents.qmix.qmix import DEFAULT_CONFIG
 from ray.rllib.policy.rnn_sequencing import chop_into_sequences
 
-from marllib.marl.models.zoo.jointQ_rnn import JointQ_RNN
-from marllib.marl.models.zoo.mixers import QMixer, VDNMixer
-from marllib.marl.algos.utils.episode_execution_plan import episode_execution_plan
+from marllib.marl.models.zoo.rnn.jointQ_rnn import JointQRNN
+from marllib.marl.models.zoo.mlp.jointQ_mlp import JointQMLP
 
+from marllib.marl.models.zoo.mixer import QMixer, VDNMixer
+from marllib.marl.algos.utils.episode_execution_plan import episode_execution_plan
 
 # original _unroll_mac for next observation is different from Pymarl.
 # thus we provide a new JointQLoss here
@@ -194,6 +212,7 @@ class JointQPolicy(Policy):
             self.obs_size = _get_size(agent_obs_space)
             self.env_global_state_shape = (self.obs_size, self.n_agents)
 
+        core_arch = config["model"]["custom_model_config"]["model_arch_args"]["core_arch"]
         self.model = ModelCatalog.get_model_v2(
             agent_obs_space,
             action_space.spaces[0],
@@ -201,7 +220,7 @@ class JointQPolicy(Policy):
             config["model"],
             framework="torch",
             name="model",
-            default_model=JointQ_RNN).to(self.device)
+            default_model=JointQMLP if core_arch == "mlp" else JointQRNN).to(self.device)
 
         self.target_model = ModelCatalog.get_model_v2(
             agent_obs_space,
@@ -210,7 +229,7 @@ class JointQPolicy(Policy):
             config["model"],
             framework="torch",
             name="target_model",
-            default_model=JointQ_RNN).to(self.device)
+            default_model=JointQMLP if core_arch == "mlp" else JointQRNN).to(self.device)
 
         self.exploration = self._create_exploration()
 
@@ -219,7 +238,7 @@ class JointQPolicy(Policy):
         if custom_config["global_state_flag"]:
             state_dim = custom_config["space_obs"]["state"].shape
         else:
-            state_dim = custom_config["space_obs"]["obs"].shape + (custom_config["num_agents"], )
+            state_dim = custom_config["space_obs"]["obs"].shape + (custom_config["num_agents"],)
         if config["mixer"] is None:  # "iql"
             self.mixer = None
             self.target_mixer = None
@@ -514,7 +533,6 @@ class JointQPolicy(Policy):
         return obs, action_mask, state
 
 
-
 JointQTrainer = GenericOffPolicyTrainer.with_updates(
     name="JointQ",
     default_config=DEFAULT_CONFIG,
